#==================================================================================================
# Imports and defines
#==================================================================================================
import bpy
import mathutils
import math
import re
import json
from mathutils import Vector, Matrix, Quaternion,Euler
from bpy.props import (StringProperty,
                       BoolProperty,
                       IntProperty,
                       FloatProperty,
                       EnumProperty,
                       PointerProperty,
                       CollectionProperty,
                       FloatVectorProperty
                       )
from bpy.types import (Panel,
                       Operator,
                       PropertyGroup,
                       Header
                       )
from math import degrees  
from os import path
#==================================================================================================
# 
#==================================================================================================






#==================================================================================================
# Addon Info
#==================================================================================================
bl_info = {
    "name": "FaceRigger",
    "description": "Rigs a model's face using either shape keys or face bones via Pose Library/Actions",
    "author": "Daitomodachi",
    "version": (1, 0, 0),
    "blender": (2, 82, 0),
    "location": "Object Panel",
    "warning": "",
    "wiki_url": "",
    "tracker_url": "",
    "support": "COMMUNITY",
    "category": "Rigging"
}
#==================================================================================================
# 
#==================================================================================================    






#==================================================================================================
# Global variables used in the Face Rigger code
#==================================================================================================

# Variables used for the appending the Face Rigger components
blend_file = path.join(path.dirname(__file__), "./FaceRigger Components.blend")   
section = "\\Collection\\"
obj = "Face Rigger Components"

# Variable used for storing the location of the JSON file containing the words used for the prefill option
Face_Flexes_Json_File = path.join(path.dirname(__file__), "./FaceRig Flex.json")


# Gets the JSON file storing the Flexes info
with open(Face_Flexes_Json_File) as f:
     Face_Flexes = json.load(f)

#Global Variables
# Arm_Name = 'Armature'
# Head = bpy.context.active_object

# Properties array used for the prefill function
FaceRig_Properties = ['Look_Left_SK', 'Look_Right_SK', 'Look_Up_SK', 'Look_Down_SK','Left_Blink','Right_Blink',
'Left_Wide','Right_Wide','Left_Squint','Right_Squint','Left_Middle_Eyebrow_Up','Right_Middle_Eyebrow_Up',
'Left_Middle_Eyebrow_Down','Right_Middle_Eyebrow_Down','Left_Inner_Eyebrow_Up','Right_Inner_Eyebrow_Up',
'Left_Inner_Eyebrow_Down','Right_Inner_Eyebrow_Down','Left_Lipcorner_Up','Right_Lipcorner_Up','Left_Lipcorner_Down','Right_Lipcorner_Down',
'Phoneme_Ah','Phoneme_Oh','Phoneme_Tn','Phoneme_Bm','Phoneme_Fv']

#==================================================================================================
# 
#==================================================================================================






#==================================================================================================
# Sets the bone layer of the armature the code is currently in
#==================================================================================================
def set_bone_layer(layer,arm_name):
    
    #Creates a layers list and set the layer that will be present
    new_layers = [False] * 32
    new_layers[layer] = True

    # Gets the name of the current Armature
    bones = bpy.data.objects[arm_name].data.bones

    # Iterates through the bones setting it to the new layer
    for i in bones: i.layers = new_layers
 
#==================================================================================================
# 
#==================================================================================================

   



    
#==================================================================================================
# Find Left and Right Bone functions
#==================================================================================================
def find_left_bone(name):
    # Sets the lists of prefixes and sufixes 
    l_prefixes = ['L','l','L_','l_','L.','l.','Left','left']
    l_sufixes = [' L',' l','_L','_l','.L','.l','Left','left']
    l_middle = [' Left ',' left ']

    # Iterates through the suffixes, prefixes, and middle list. If it finds any of these words inside name, then it returns true
    for l in l_prefixes:
        if name.startswith(l):return True
    for l in l_sufixes:
        if name.endswith(l):return True
    for l in l_middle:
        if name.find(l) != -1:return True
    # If it cannot find any of these names, then it returns false
    return False           

def find_right_bone(name):
    # Sets the lists of prefixes and sufixes 
    r_prefixes = ['R','r','R_','r_','R.','r.','Right','right']
    r_sufixes = [' R',' r','_R','_r','.R','.r','Right','right']
    r_middle = [' Right ',' right ']


    # Iterates through the suffixes, prefixes, and middle list. If it finds any of these words inside name, then it returns true
    for r in r_prefixes:
        if name.startswith(r):return True
    for r in r_sufixes:
        if name.endswith(r):return True
    for r in r_middle:
        if name.find(r) != -1: return True
     # If it cannot find any of these names, then it returns false
    return False     
#==================================================================================================
# 
#==================================================================================================






#==================================================================================================
# Remove Bone Function
#==================================================================================================      
def remove_bone_from_action(bone_name, action):
    # Gets the action curves
    fcu = action.fcurves
    channels = ['location', 'rotation','rotation_euler', 'rotation_quaternion', 'scale']

    # Iterates through the channels of the curve
    for attr in channels:
        for i in range(0, 4):
            curve = fcu.find(f'pose.bones["{bone_name}"].{attr}', index=i)
            if curve is None: continue
            fcu.remove(curve)
#==================================================================================================
# 
#==================================================================================================
                
 



#==================================================================================================
# Split Action Function
#================================================================================================== 
def split_action(action_name):
    
    # Creates a left action strip from the action
    action = bpy.data.actions[action_name]
    action_copy = action.copy()
    action_copy.name = '%s L'%(action_name)
    action_fcu = action_copy.fcurves
    bones = set()
    ob = bpy.context.object
    
    # Gets all the bones from the action
    for fcu in action_fcu:
        bone = fcu.data_path.split('"')[1]
        bones.add(bone)

    
    for i in bones:
        pt = ob.pose.bones[i].name
         # Checks to see if .r is in the bone's name
        if(find_right_bone(pt) is True):
            # Removes the bone from the action
            remove_bone_from_action(i, action_copy)

   
    #Creates a right stripe from the action
    action2 = bpy.data.actions[action_name]       
    action2_copy = action.copy()
    action2_copy.name = '%s R'%(action_name)
    action2_fcu = action_copy.fcurves
    bones = set()
    ob = bpy.context.object
    
    # Like before, it gets all the bones from the action
    for fcu in action2_fcu:
        bone = fcu.data_path.split('"')[1]
        bones.add(bone)
    for i in bones:
        pt = ob.pose.bones[i].name
         # Checks to see if .r is in the bone's name
        if(find_left_bone(pt) is True):
            # Removes the bone from the action
            remove_bone_from_action(i, action2_copy)
 #==================================================================================================
# 
#==================================================================================================





           
#==================================================================================================
# Add Action Constraint
#==================================================================================================           
def add_action_constraint(driver_bone, driver_obj,action, con_axis='X', con_min=0, con_max=1.0):
# Checks to see if the user wants the action split    
    Action = bpy.data.actions[action]   
    bones = set()
    
    # Retrieves all of the bones from the action
    for fcu in Action.fcurves:
        bone = fcu.data_path.split('"')[1]
        bones.add(bone)
    
    # Iterates through all of the bones and adds an action constraint to them
    for i in bones:
        con = driver_obj.pose.bones[i].constraints.new('ACTION')
        con.target = driver_obj
        con.subtarget = driver_bone.name
        con.target_space = 'LOCAL'
        con.transform_channel = f'LOCATION_{con_axis}'
        con.action = bpy.data.actions[action]
        con.frame_start = int(Action.frame_range[0])
        con.frame_end = int(Action.frame_range[1])
        con.min = con_min
        con.max = con_max
#==================================================================================================
# 
#==================================================================================================






#==================================================================================================
# Prefill function
#==================================================================================================     
def prefill(self,context):
    # Runs as long as the selected mesh has shape keys attached 
    if bpy.data.objects[bpy.context.scene.Face.name].data.shape_keys is not None:
        # retrieves the face rig data object
        data = bpy.context.scene.Face.face_rig_data 
        
        # sets a variable to store the shape keys
        shape_keys = bpy.data.objects[bpy.context.scene.Face.name].data.shape_keys.key_blocks
        
        for f in FaceRig_Properties:
            # Iterates through the array located at the indicated key
            for i in Face_Flexes[f]:
                # Checks to see if the name matches up with a shape key in the mesh
                if shape_keys.find(i) is not -1:
                    # Sets the shape key name at the indicated property and exits out of the function
                    setattr(data,f, i)
                    break
#==================================================================================================
# 
#==================================================================================================                    
        





#==================================================================================================
# FaceRigData Class
#==================================================================================================
class FaceRigData(bpy.types.PropertyGroup):
    
    # Determines if the the tables and facerigger should be set to the aramature. Ideal for ARP rigs
    Join_To_Armature : bpy.props.BoolProperty(name="Join To Armature")
    
    # Select a mode
    Shape_Keys : bpy.props.BoolProperty(name="Use Shape Keys")
    Face_Bones : bpy.props.BoolProperty(name="Use Face Bones")


# Properties uses for the Rigger if the user selects the Shape Key setting

    # Eye Target Properties
    Eye : bpy.props.StringProperty(name="Eye")
    Look_Left_SK : bpy.props.StringProperty(name="Look_Left_SK" )
    Look_Right_SK : bpy.props.StringProperty(name="Look_Right_SK")
    Look_Up_SK : bpy.props.StringProperty(name="Look_Up_SK")
    Look_Down_SK : bpy.props.StringProperty(name="Look_Down_SK")
    
    # Upper Eyelid Properties
    Left_Blink : bpy.props.StringProperty(name="Left_Blink")
    Right_Blink : bpy.props.StringProperty(name="Right_Blink")
    Left_Wide : bpy.props.StringProperty(name="Left_Wide")
    Right_Wide : bpy.props.StringProperty(name="Right_Wide")
    
    # Lower Eyelid Properties
    Left_Squint : bpy.props.StringProperty(name="Left_Squint")
    Right_Squint : bpy.props.StringProperty(name="Right_Squint")

    # Eyebrow Properties
    Left_Middle_Eyebrow_Up : bpy.props.StringProperty(name="Left_Middle_Eyebrow_Up")
    Right_Middle_Eyebrow_Up : bpy.props.StringProperty(name="Right_Middle_Eyebrow_Up")
    Left_Middle_Eyebrow_Down : bpy.props.StringProperty(name="Left_Middle_Eyebrow_Down")
    Right_Middle_Eyebrow_Down : bpy.props.StringProperty(name="Right_Middle_Eyebrow_Down")
    Left_Inner_Eyebrow_Up : bpy.props.StringProperty(name="Left_Inner_Eyebrow_Up")
    Right_Inner_Eyebrow_Up : bpy.props.StringProperty(name="Right_Inner_Eyebrow_Up")
    Left_Inner_Eyebrow_Down : bpy.props.StringProperty(name="Left_Inner_Eyebrow_Down")
    Right_Inner_Eyebrow_Down : bpy.props.StringProperty(name="Right_Inner_Eyebrow_Down")
    
    # Lip Corners
    Left_Lipcorner_Up : bpy.props.StringProperty(name="Left_Lipcorner_Up")
    Right_Lipcorner_Up : bpy.props.StringProperty(name="Right_Lipcorner_Up")
    Left_Lipcorner_Down : bpy.props.StringProperty(name="Left_Lipcorner_Down")
    Right_Lipcorner_Down : bpy.props.StringProperty(name="Right_Lipcorner_Down")
    
    #Phoneme Library
    Phoneme_Ah : bpy.props.StringProperty(name="Phoneme_Ah")
    Phoneme_Oh : bpy.props.StringProperty(name="Phoneme_Oh")
    Phoneme_Tn : bpy.props.StringProperty(name="Phoneme_Tn")
    Phoneme_Bm : bpy.props.StringProperty(name="Phoneme_Bm")
    Phoneme_Fv : bpy.props.StringProperty(name="Phoneme_Fv")


# Properties uses for the Rigger if the user selects the Face Bones setting
    Eyes_UpDown    : bpy.props.StringProperty(name="Eyes_UpDown")
    Eyes_LeftRight : bpy.props.StringProperty(name="Eyes_LeftRight")
    Closedlid            : bpy.props.StringProperty(name="Closedlid")
    Lowerlid             : bpy.props.StringProperty(name="Lowerlid")
    BrowIn               : bpy.props.StringProperty(name="BrowIn")
    Brow                 : bpy.props.StringProperty(name="Brow")
    Lipcorner_UpDown     : bpy.props.StringProperty(name="Lip Corners Up Down")
    Lipcorner_LeftRight  : bpy.props.StringProperty(name="Lip Corners Left Right")
    
    
# Variables to handle blender layers
    Phoneme_Table_Bone_Layers : bpy.props.IntProperty(name="Phoneme Table Active Layer")
    FaceRig_Bone_Layers : bpy.props.IntProperty(name="FaceRig Table Active Layer")
    
# Stores the bone name for locking the head bone to
    Tables_Parent_Bone  : bpy.props.StringProperty(name="Tables Parent Bone")

#==================================================================================================
#
#==================================================================================================






#==================================================================================================
# Clean
#==================================================================================================
def clean():
        # Clears all of the FaceRig data info (if there's a more efficent way of doing this, it would be appeaciated
        data.Look_Left_SK = ""
        data.Look_Right_SK = ""
        data.Look_Up_SK = ""
        data.Look_Down_SK = ""
        data.Left_Blink = ""
        data.Right_Blink = ""
        data.Left_Middle_Eyebrow_Up = ""
        data.Left_Middle_Eyebrow_Down = ""
        data.Right_Middle_Eyebrow_Up = ""
        data.Right_Middle_Eyebrow_Down = ""
        data.Right_Inner_Eyebrow_Up = ""
        data.Left_Inner_Eyebrow_Up = ""
        data.Right_Inner_Eyebrow_Down = ""
        data.Left_Inner_Eyebrow_Down = ""
        data.Left_Lipcorner_Up = ""
        data.Left_Lipcorner_Down = ""
        data.Right_Lipcorner_Up = ""
        data.Right_Lipcorner_Down = ""
        data.Phoneme_Ah = ""
        data.Phoneme_Oh = ""
        data.Phoneme_Tn = ""
        data.Phoneme_Bm = ""
        data.Phoneme_Fv = ""
        data.Left_Wide = ""
        data.Right_Wide = ""
        data.Left_Squint = ""
        data.Right_Squint = ""
        data.Eyes_UpDown = ""
        data.Eyes_LeftRight = ""
        data.Closedlid = ""
        data.Lowerlid = ""
        data.BrowIn = ""
        data.Brow = ""
        data.Lipcorner_UpDown = ""
        data.Lipcorner_LeftRight = ""

        data.Shape_Keys = False
        data.Face_Bones = False

        data.Phoneme_Table_Bone_Layers = 0
        data.FaceRig_Bone_Layers = 0
        bpy.context.scene.Face = None
        data.Tables_Parent_Bone = ""
#==================================================================================================
# 
#==================================================================================================






#==================================================================================================
# Builds FaceRig witg Shape Keys
#==================================================================================================      
def Build_FaceRig_shape_key(Obj):

        # Sets the active object and selects the face rig         
        bpy.context.view_layer.objects.active = Obj
        FaceRig.select_set(True) 
        Obj.select_set(True)

        # Joins the Face Rig and the Object together
        bpy.ops.object.join()

        # Go into object mode
        bpy.ops.object.mode_set(mode='POSE')

        # Sets up the Eye Target Constraints
        if data.Look_Left_SK  is not "" : add_driver_FaceRig(data.Look_Left_SK, 'FaceRig.EyeTarget',"L")
        if data.Look_Right_SK is not "" : add_driver_FaceRig(data.Look_Right_SK, 'FaceRig.EyeTarget',"R")
        if data.Look_Down_SK  is not "" : add_driver_FaceRig(data.Look_Down_SK, 'FaceRig.EyeTarget',"D")
        if data.Look_Up_SK    is not "" : add_driver_FaceRig(data.Look_Up_SK, 'FaceRig.EyeTarget',"U")
        add_limit_FaceRig_shape_keys('FaceRig.EyeTarget', "A")
        
        # Sets up the Upper Eyelid Controls
        if data.Right_Blink  is not "" : add_driver_FaceRig(data.Right_Blink, 'FaceRig.UpperEyelid.Slider.R',"D")
        if data.Left_Blink  is not "" : add_driver_FaceRig(data.Left_Blink, 'FaceRig.UpperEyelid.Slider.L',"D")
        if data.Right_Wide  is not "" : add_driver_FaceRig(data.Right_Wide, 'FaceRig.UpperEyelid.Slider.R',"U")
        if data.Left_Wide  is not "" : add_driver_FaceRig(data.Left_Wide, 'FaceRig.UpperEyelid.Slider.L',"U")
        add_limit_FaceRig_shape_keys('FaceRig.UpperEyelid.Slider.R', "V")
        add_limit_FaceRig_shape_keys('FaceRig.UpperEyelid.Slider.L', "V")
        
        #Sets up the Lower Eyelid Controls
        if data.Left_Squint  is not "" : add_driver_FaceRig(data.Left_Squint, 'FaceRig.LowerEyelid.Slider.R',"U")
        if data.Right_Squint  is not "" : add_driver_FaceRig(data.Right_Squint, 'FaceRig.LowerEyelid.Slider.L',"U")
        add_limit_FaceRig_shape_keys('FaceRig.LowerEyelid.Slider.L', "V")
        add_limit_FaceRig_shape_keys('FaceRig.LowerEyelid.Slider.R', "V")
        
        # Sets up the Middle Eyebrow Drivers
        if data.Left_Middle_Eyebrow_Up  is not "" : add_driver_FaceRig(data.Left_Middle_Eyebrow_Up, 'FaceRig.MiddleEyeBrow.Slider.L',"U")
        if data.Right_Middle_Eyebrow_Up  is not "" : add_driver_FaceRig(data.Right_Middle_Eyebrow_Up, 'FaceRig.MiddleEyeBrow.Slider.R',"U")
        if data.Left_Middle_Eyebrow_Down  is not "" : add_driver_FaceRig(data.Left_Middle_Eyebrow_Down, 'FaceRig.MiddleEyeBrow.Slider.L',"D")
        if data.Right_Middle_Eyebrow_Down  is not "" : add_driver_FaceRig(data.Right_Middle_Eyebrow_Down, 'FaceRig.MiddleEyeBrow.Slider.R',"D")
        add_limit_FaceRig_shape_keys('FaceRig.MiddleEyeBrow.Slider.L', "V")
        add_limit_FaceRig_shape_keys('FaceRig.MiddleEyeBrow.Slider.R', "V")    

        #Sets up the Inner Eyebrow Drivers
        if data.Left_Inner_Eyebrow_Up  is not "" : add_driver_FaceRig(data.Left_Inner_Eyebrow_Up, 'FaceRig.InnerEyeBrow.Slider.L',"U")
        if data.Right_Inner_Eyebrow_Up  is not "" : add_driver_FaceRig(data.Right_Inner_Eyebrow_Up, 'FaceRig.InnerEyeBrow.Slider.R',"U")
        if data.Left_Inner_Eyebrow_Down  is not "" : add_driver_FaceRig(data.Left_Inner_Eyebrow_Down, 'FaceRig.InnerEyeBrow.Slider.L',"D")
        if data.Right_Inner_Eyebrow_Down  is not "" : add_driver_FaceRig(data.Right_Inner_Eyebrow_Down, 'FaceRig.InnerEyeBrow.Slider.R',"D")
        add_limit_FaceRig_shape_keys('FaceRig.InnerEyeBrow.Slider.L', "V")
        add_limit_FaceRig_shape_keys('FaceRig.InnerEyeBrow.Slider.R', "V")
        
        # Sets up the lipcorner drivers
        if data.Left_Lipcorner_Up  is not "" : add_driver_FaceRig(data.Left_Lipcorner_Up, 'FaceRig.LipCorner.Target.L',"U")
        if data.Right_Lipcorner_Up  is not "" : add_driver_FaceRig(data.Right_Lipcorner_Up, 'FaceRig.LipCorner.Target.R',"U")
        if data.Left_Lipcorner_Down  is not "" : add_driver_FaceRig(data.Left_Lipcorner_Down, 'FaceRig.LipCorner.Target.L',"D")
        if data.Right_Lipcorner_Down  is not "" : add_driver_FaceRig(data.Right_Lipcorner_Down, 'FaceRig.LipCorner.Target.R',"D")
        add_limit_FaceRig_shape_keys('FaceRig.LipCorner.Target.L', "V")
        add_limit_FaceRig_shape_keys('FaceRig.LipCorner.Target.R', "V")
        

        # Goes back into Object mode
        bpy.ops.object.mode_set(mode='OBJECT')
#==================================================================================================
# 
#==================================================================================================






#==================================================================================================
# Builds the FaceRig using Action Constaints
#==================================================================================================        
def Build_FaceRig_action_constraint(Obj):
        # Stores a dictionary of actions
        FaceRig_Actions = {
        data.Closedlid : ['FaceRig.UpperEyelid.Slider.R' ,'FaceRig.UpperEyelid.Slider.L'],
        data.BrowIn : ['FaceRig.InnerEyeBrow.Slider.R' ,'FaceRig.InnerEyeBrow.Slider.L'],
        data.Brow : ['FaceRig.MiddleEyeBrow.Slider.R' ,'FaceRig.MiddleEyeBrow.Slider.L'],
        data.Lowerlid: ['FaceRig.LowerEyelid.Slider.R' ,'FaceRig.LowerEyelid.Slider.L'],
        data.Lipcorner_UpDown: ['FaceRig.LipCorner.Target.R','Y','FaceRig.LipCorner.Target.L','Y'],
        data.Lipcorner_LeftRight: ['FaceRig.LipCorner.Target.R','X','FaceRig.LipCorner.Target.L','X']
        }
    
    
    
        # Creates 3 dictionaires: one for action constriants constrainted to the x, one constrained to the y, and one dictionary the defines the limit locations
        limit_location_constraints = {
        'FaceRig.EyeTarget': 'A',
        'FaceRig.UpperEyelid.Slider.R' : 'V', 
        'FaceRig.UpperEyelid.Slider.L' : 'V',
        'FaceRig.LowerEyelid.Slider.L' : 'V',
        'FaceRig.LowerEyelid.Slider.R' : 'V',
        'FaceRig.MiddleEyeBrow.Slider.R' : 'V',
        'FaceRig.MiddleEyeBrow.Slider.L' : 'V',
        'FaceRig.InnerEyeBrow.Slider.R' : 'V',
        'FaceRig.InnerEyeBrow.Slider.L' : 'V',
        'FaceRig.LipCorner.Target.R' : 'A',
        'FaceRig.LipCorner.Target.L' : 'A'
        
        }
        
        
        
        # Stores the object 
        Obj = bpy.data.objects[Arm_Name]
        
        # Selects the object and the face rig
        bpy.context.view_layer.objects.active = Obj
        FaceRig.select_set(True) 
        Obj.select_set(True) 

        # Joins the 2 objects together
        bpy.ops.object.join()

        # Go into pose mode
        bpy.ops.object.mode_set(mode='POSE')
        
        # Gets all of the keys from the FaceRig Actions dictionary
        keys = FaceRig_Actions.keys()
        for k in keys:
            
            # Splits the current action into a left and right part depending on the location of the k in the for loop
            split_action(k)            
            for i in FaceRig_Actions[k]:
                # Checks to see if the current facerig bone is a left bone
                if i.find('.L') != -1 and i.endswith('.L') is True:
                
                    # Creates an action name with a L at the end
                    Left_Action_Name = '%s L'%(k)
                    Current_Bone = Obj.pose.bones[i]
                    
                    # Checks to see if the current bone is an target bone
                    if i.find('.Target.') != -1:
                    
                        # Checks to see if it's for horizontal movement. This is necessary since the X axis is flipped around
                        if (FaceRig_Actions[k][FaceRig_Actions[k].index(i)+1]) == 'X':
                            new_min = 0.2*FaceRig_Scale
                            new_max = -0.2*FaceRig_Scale
                        else:
                            new_min = -0.2*FaceRig_Scale
                            new_max = 0.2*FaceRig_Scale
                        
                        # Adds the action constraint
                        add_action_constraint(Current_Bone, Obj,Left_Action_Name, con_axis=FaceRig_Actions[k][FaceRig_Actions[k].index(i)+1], con_min=new_min, con_max=new_max)
                    else:
                        # Adds the constraint
                        add_action_constraint(Current_Bone, Obj,Left_Action_Name, con_axis='Y', con_min=-0.2*FaceRig_Scale, con_max=0.2*FaceRig_Scale)
                        
                # Checks to see if the bone is a right bone
                elif i.find('.R') != -1 and i.endswith('.R') is True:
                    
                    Right_Action_Name = '%s R'%(k)
                    Current_Bone = Obj.pose.bones[i]
                    
                    if i.find('.Target.') != -1:
                        add_action_constraint(Current_Bone, Obj,Right_Action_Name, con_axis=FaceRig_Actions[k][FaceRig_Actions[k].index(i)+1], con_min=-0.2*FaceRig_Scale, con_max=0.2*FaceRig_Scale)
                    else:
                        add_action_constraint(Current_Bone, Obj,Right_Action_Name, con_axis='Y', con_min=-0.2*FaceRig_Scale, con_max=0.2*FaceRig_Scale)
                  # Gets the keys from the y dictionary
                  
        # Eye target and lip corner targets are the only FaceRigger bone that gets an explicit case
        Current_Bone = Obj.pose.bones['FaceRig.EyeTarget']
        if data.Eyes_UpDown     is not "" : add_action_constraint(Current_Bone, Obj,data.Eyes_UpDown, con_axis='Y', con_min=-0.2*FaceRig_Scale, con_max=0.2*FaceRig_Scale)
        if data.Eyes_LeftRight  is not "" : add_action_constraint(Current_Bone, Obj,data.Eyes_LeftRight, con_axis='X', con_min=-0.2*FaceRig_Scale, con_max=0.2*FaceRig_Scale)     
        
   
  
        # Gets all of the keys from the location dictionary
        keys = limit_location_constraints.keys()
        for k in keys:
            if limit_location_constraints[k] is not "" : add_limit_FaceRig_shape_keys(k, limit_location_constraints[k])
        
        # Goes into object mode
        bpy.ops.object.mode_set(mode='OBJECT')     
#==================================================================================================
# 
#==================================================================================================        






#==================================================================================================
# Builds a Phoneme Table using Action Constraints
#==================================================================================================        
def Build_Phoneme_action_constraint(Obj):        
        
        # Stores the object 
        Obj = bpy.data.objects[Arm_Name]
        
        
        bpy.context.view_layer.objects.active = Obj
        Phoneme_Table.select_set(True)
        Obj.select_set(True)

        bpy.ops.object.join()

        bpy.ops.object.mode_set(mode='POSE')
        
        Current_Bone =Obj.pose.bones['Phoneme Table.Ah']
        if data.Phoneme_Ah     is not "" : add_action_constraint(Current_Bone, Obj,data.Phoneme_Ah, con_axis='Y', con_min=0, con_max=1.16244*Phoneme_Table_Scale)
        add_limit_Phoneme_Table_shape_keys('Phoneme Table.Ah') if data.Phoneme_Ah is not "" else lock_bone('Phoneme Table.Ah')
        
        Current_Bone =Obj.pose.bones['Phoneme Table.Oh']
        if data.Phoneme_Oh     is not "" : add_action_constraint(Current_Bone, Obj,data.Phoneme_Oh, con_axis='Y', con_min=0, con_max=1.16244*Phoneme_Table_Scale)
        add_limit_Phoneme_Table_shape_keys('Phoneme Table.Oh') if data.Phoneme_Oh is not "" else lock_bone('Phoneme Table.Oh')
        
          
        Current_Bone =Obj.pose.bones['Phoneme Table.Tn']
        if data.Phoneme_Tn     is not "" : add_action_constraint(Current_Bone, Obj,data.Phoneme_Tn, con_axis='Y', con_min=0, con_max=1.16244*Phoneme_Table_Scale)
        add_limit_Phoneme_Table_shape_keys('Phoneme Table.Tn') if data.Phoneme_Tn is not "" else lock_bone('Phoneme Table.Tn')
 
          
        Current_Bone =Obj.pose.bones['Phoneme Table.BM']
        if data.Phoneme_Bm is not "" : add_action_constraint(Current_Bone, Obj,data.Phoneme_Bm, con_axis='Y', con_min=0, con_max=1.16244*Phoneme_Table_Scale)
        add_limit_Phoneme_Table_shape_keys('Phoneme Table.BM') if data.Phoneme_Bm is not "" else lock_bone('Phoneme Table.BM')
 
          
        Current_Bone =Obj.pose.bones['Phoneme Table.FV']
        if data.Phoneme_Fv is not "" : add_action_constraint(Current_Bone, Obj,data.Phoneme_Fv, con_axis='Y', con_min=0, con_max=1.16244*Phoneme_Table_Scale)
        add_limit_Phoneme_Table_shape_keys('Phoneme Table.FV') if data.Phoneme_Fv is not "" else lock_bone('Phoneme Table.FV')
 
        bpy.ops.object.mode_set(mode='OBJECT')
#==================================================================================================
# 
#==================================================================================================   








#==================================================================================================
# Add Limit To FaceRig Shape Keys
#==================================================================================================   
def add_limit_FaceRig_shape_keys(Driver_Bone_Name, Direction):

    # Sets the newly created controller bone as the bone we'll be setting as our target
    selected_bone = bpy.data.objects[Arm_Name].pose.bones[Driver_Bone_Name]
    

    # Creates a limit location constraint
    constraints = selected_bone.constraints
    limit_con = constraints.new('LIMIT_LOCATION')
    
    if Direction == "V":
        limit_con.owner_space = 'LOCAL'
        limit_con.use_min_y = True
        limit_con.min_y = -.2 * FaceRig_Scale
        limit_con.use_max_y = True
        limit_con.max_y = .2 * FaceRig_Scale

        limit_con.use_transform_limit = True
        limit_con.use_min_x = True
        limit_con.min_x = 0
        limit_con.use_max_x = True
        limit_con.max_x = 0
    elif Direction == "H":
        limit_con.owner_space = 'LOCAL'
        limit_con.use_min_x = True
        limit_con.min_x = -.2 * FaceRig_Scale
        limit_con.use_max_x = True
        limit_con.max_x = .2 * FaceRig_Scale

        limit_con.use_transform_limit = True
        limit_con.use_min_y = True
        limit_con.min_y = 0
        limit_con.use_max_y = True
        limit_con.max_y = 0
    else:
        limit_con.owner_space = 'LOCAL'
        limit_con.use_min_x = True
        limit_con.min_x = -.2 * FaceRig_Scale
        limit_con.use_max_x = True
        limit_con.max_x = .2 * FaceRig_Scale

        limit_con.use_transform_limit = True
        limit_con.use_min_y = True
        limit_con.min_y = -.2 * FaceRig_Scale
        limit_con.use_max_y = True
        limit_con.max_y = .2 * FaceRig_Scale

#==================================================================================================
# 
#==================================================================================================   





 

#==================================================================================================
# Adds limit to Phoneme Table With ShapeKeys
#==================================================================================================   
def add_limit_Phoneme_Table_shape_keys(Driver_Bone_Name):

    # Sets the newly created controller bone as the bone we'll be setting as our target
    selected_bone = bpy.data.objects[Arm_Name].pose.bones[Driver_Bone_Name]
    

    # Creates a limit location constraint
    constraints = selected_bone.constraints
    limit_con = constraints.new('LIMIT_LOCATION')
    
    limit_con.owner_space = 'LOCAL'
    limit_con.use_min_y = True
    limit_con.min_y = 0
    limit_con.use_max_y = True
    limit_con.max_y = 1.16244 * Phoneme_Table_Scale

    limit_con.use_transform_limit = True
    limit_con.use_min_x = True
    limit_con.min_x = 0
    limit_con.use_max_x = True
    limit_con.max_x = 0
#==================================================================================================
# 
#==================================================================================================       




 

#==================================================================================================
# Builds Phoneme Table with Shape Keys
#==================================================================================================    
def Build_Phoneme_shape_key(Obj):
        
        # Selects the Table and the object
        bpy.context.view_layer.objects.active = Obj
        Phoneme_Table.select_set(True)
        Obj.select_set(True)
        
        # Joins the two objects
        bpy.ops.object.join()

        # Goes into Pose mode
        bpy.ops.object.mode_set(mode='POSE')


        
        # Sets up the Ah constraints
        if data.Phoneme_Ah is not "" : add_driver_Phoneme_Table(data.Phoneme_Ah, 'Phoneme Table.Ah')
        add_limit_Phoneme_Table_shape_keys('Phoneme Table.Ah') if data.Phoneme_Ah is not "" else lock_bone('Phoneme Table.Ah')
        
        # Sets up the Ah constraints
        if data.Phoneme_Oh is not "" : add_driver_Phoneme_Table(data.Phoneme_Oh, 'Phoneme Table.Oh')
        add_limit_Phoneme_Table_shape_keys('Phoneme Table.Oh') if data.Phoneme_Oh is not "" else lock_bone('Phoneme Table.Oh')
        
        # Sets up the Ah constraints
        if data.Phoneme_Tn is not "" : add_driver_Phoneme_Table(data.Phoneme_Tn, 'Phoneme Table.Tn')
        add_limit_Phoneme_Table_shape_keys('Phoneme Table.Tn') if data.Phoneme_Tn is not "" else lock_bone('Phoneme Table.Tn')
        
        # Sets up the Ah constraints
        if data.Phoneme_Bm is not "" : add_driver_Phoneme_Table(data.Phoneme_Bm, 'Phoneme Table.BM')
        add_limit_Phoneme_Table_shape_keys('Phoneme Table.BM') if data.Phoneme_Bm is not "" else lock_bone('Phoneme Table.BM')
        
        # Sets up the Ah constraints
        if data.Phoneme_Fv is not "" : add_driver_Phoneme_Table(data.Phoneme_Fv, 'Phoneme Table.FV')
        add_limit_Phoneme_Table_shape_keys('Phoneme Table.FV') if data.Phoneme_Fv is not "" else lock_bone('Phoneme Table.FV')
        

        # Goes back into object mode
        bpy.ops.object.mode_set(mode='OBJECT')

#==================================================================================================
# 
#==================================================================================================   





#==================================================================================================
# Locks the current bone preventing it from movement
#==================================================================================================   
def lock_bone(Bone):
    bpy.data.objects[Arm_Name].pose.bones[Bone].lock_location = (True,True,True)
#==================================================================================================
#
#==================================================================================================   







#==================================================================================================
# Add driver to Face Rig
#==================================================================================================      
# Adds a driver to the desired Driver Bone. 
def add_driver_FaceRig(Shape_Key_Name, Driver_Bone_Name, Direction):
    # Sets the newly created controller bone as the bone we'll be setting as our target
    selected_bone = bpy.data.objects[Arm_Name].pose.bones[Driver_Bone_Name]

    # Sets up the basic parameters for setting up the driver
    Head_Shape_Keys = Head.data.shape_keys
    Head_Desired_Shape_Key =Head_Shape_Keys.key_blocks[Shape_Key_Name]
    drv = Head_Desired_Shape_Key.driver_add('value').driver
    var = drv.variables.new()
    var.name = 'var'
    var.type = 'TRANSFORMS'
    drv.type = 'SCRIPTED'
    target = var.targets[0]
    target.id = bpy.data.objects[Arm_Name]
    target.bone_target = Driver_Bone_Name
    target.transform_space = 'LOCAL_SPACE'
    
    if Direction == "L":
        target.transform_type = 'LOC_X'
        drv.expression = f"var*(1/{.2*FaceRig_Scale:f})"
    elif Direction == "R":
        target.transform_type = 'LOC_X'
        drv.expression = f"var*(-1/{.2*FaceRig_Scale:f})"
    elif Direction == "U":
        target.transform_type = 'LOC_Y'
        drv.expression = f"var*(1/{.2*FaceRig_Scale:f})"
    else:
        target.transform_type = 'LOC_Y'
        drv.expression = f"var*(-1/{.2*FaceRig_Scale:f})"
#==================================================================================================
# 
#================================================================================================== 






#==================================================================================================
# Adds a driver to the Phoneme Table
#==================================================================================================   
def add_driver_Phoneme_Table(Shape_Key_Name, Driver_Bone_Name):
    # Sets the newly created controller bone as the bone we'll be setting as our target
    selected_bone = bpy.data.objects[Arm_Name].pose.bones[Driver_Bone_Name]

    Head_Shape_Keys = Head.data.shape_keys
    Head_Desired_Shape_Key =Head_Shape_Keys.key_blocks[Shape_Key_Name]
    drv = Head_Desired_Shape_Key.driver_add('value').driver
    var = drv.variables.new()
    var.name = 'var'
    var.type = 'TRANSFORMS'
    drv.type = 'SCRIPTED'
    target = var.targets[0]
    target.id = bpy.data.objects[Arm_Name]
    target.bone_target = Driver_Bone_Name
    target.transform_space = 'LOCAL_SPACE'
    target.transform_type = 'LOC_Y'
    drv.expression = f"var*(1/{1.16244*Phoneme_Table_Scale:f})"
#==================================================================================================
# 
#================================================================================================== 






#==================================================================================================
# Loads The FaceRigger Components to the scene
#================================================================================================== 
def load_FaceRigger_Components():
    filepath = blend_file + section + obj
    directory = blend_file + section
    filename = obj
    bpy.ops.wm.append(filepath=filepath, filename=filename, directory=directory)
#==================================================================================================
# 
#================================================================================================== 






#==================================================================================================
# Load FaceRigger Compoenents Operator
#==================================================================================================  
class Load_FaceRigger_Components_Operator(bpy.types.Operator):
    """Tooltip"""
    bl_idname = "bone.load_facerig"
    bl_label = "Face Rigger: Load FaceRigger Components"


    def execute(self, context):
        load_FaceRigger_Components()
        return {'FINISHED'}

def lock_FaceRiggerComponents(bone_name):
    bpy.ops.object.mode_set(mode='EDIT')
    Table = bpy.data.objects[Arm.name].data.edit_bones['Phoneme_Table.Root']
    FaceRig = bpy.data.objects[Arm.name].data.edit_bones['FaceRig.Root']
    Head = bpy.data.objects[Arm.name].data.edit_bones[bone_name]
    
    Table.parent = Head
    FaceRig.parent = Head
    bpy.ops.object.mode_set(mode='POSE')
#==================================================================================================
# 
#================================================================================================== 






#==================================================================================================
#  Lock FaceRigger components operator
#================================================================================================== 
class FaceRiggerOp_LockComponents(Operator):
    bl_idname = "bone.lockfacerigger"
    bl_label = "Lock Face Rigger Components To Bone"

    my_string: bpy.props.StringProperty(name="Head Bone")

    def execute(self, context):
        lock_FaceRiggerComponents(self.my_string)
        return {'FINISHED'}

    def invoke(self, context, event):
        global Arm, Obj
        Obj = bpy.context.object
        Arm = bpy.data.objects[Obj.name].data
        wm = context.window_manager
        return wm.invoke_props_dialog(self)
    
    def draw(self,context):
        layout = self.layout
        row = layout.row()
        row.prop_search(self,"my_string",Arm,"bones")
#==================================================================================================
# 
#================================================================================================== 






#==================================================================================================
# Simple Operator (Handles the main opeation of the FaceRigger addon and will be renamed later
#==================================================================================================         
class SimpleOperator(bpy.types.Operator):
    """Tooltip"""
    bl_idname = "object.simple_operator"
    bl_label = "Run"

    def execute(self, context):
        if (bpy.data.objects.get("FaceRig") is not None) and (bpy.data.objects.get("Phoneme Table") is not None):
                #Sets the global variables from above to what was set in the UI
            global data
            data = bpy.context.scene.Face.face_rig_data 
            global Head
            Head = bpy.context.scene.Face
            
            # Retrieves the name of the Armature
            mod = Head.modifiers
            
            global FaceRig,FaceRig_Scale
            FaceRig = bpy.data.objects['FaceRig']
            FaceRig_Scale = FaceRig.scale.x
            
            # Checks to see if the desired Bone Layer is between 0 and 31
            if data.FaceRig_Bone_Layers >= 0 and data.FaceRig_Bone_Layers < 32:
            
                # Sets the FaceRig's desired Bone Layer
                bpy.context.view_layer.objects.active = FaceRig
                FaceRig.select_set(True)
                bpy.ops.object.mode_set(mode='POSE')
                set_bone_layer(data.FaceRig_Bone_Layers,'FaceRig')
                bpy.ops.object.mode_set(mode='OBJECT')
                FaceRig.select_set(False)   
                
            else:
                
                # If the user does not enter the right number, it will cancel
                self.report({'ERROR'}, 'Desired FaceRig Bone Layer should be between 0 and 31')
                return {'FINISHED'}

            global Phoneme_Table, Phoneme_Table_Scale
            Phoneme_Table = bpy.data.objects['Phoneme Table'] 
            Phoneme_Table_Scale = Phoneme_Table.scale.x
            
            # Checks to see if the desired Bone Layer is between 0 and 31
            if data.Phoneme_Table_Bone_Layers >= 0 and data.Phoneme_Table_Bone_Layers < 32:
                           
                # Sets the Phoneme Table's desired Bone Layer
                bpy.context.view_layer.objects.active = Phoneme_Table
                Phoneme_Table.select_set(True)
                bpy.ops.object.mode_set(mode='POSE')
                set_bone_layer(data.Phoneme_Table_Bone_Layers,'Phoneme Table')
                bpy.ops.object.mode_set(mode='OBJECT')
                Phoneme_Table.select_set(False)
                
            else:    
            
                # If the user does not enter the right number, it will cancel
                self.report({'ERROR'}, 'Desired Phoneme Table Bone Layer should be between 0 and 31')
                return {'FINISHED'}
            
            if data.Shape_Keys is True:
                # Iterates through the object's modifier looking for the armature modifier
                for i in mod:
                    
                    # Runs if it finds the Armature modifier
                    if i.type == 'ARMATURE':
                        # Changes the armature name variable to the armature attached to the face
                        global Arm_Name
                        Arm_Name = i.object.name
                        
                        # Sets up the Face rig
                        Build_FaceRig_shape_key(bpy.data.objects[Arm_Name])
                        Build_Phoneme_shape_key(bpy.data.objects[Arm_Name])
                        
                        # Locks the tables to a bone if the user set it                        
                        if data.Tables_Parent_Bone != "":
                            bpy.ops.object.mode_set(mode='EDIT')
                            Table = bpy.data.objects[Arm_Name].data.edit_bones['Phoneme_Table.Root']
                            FaceRig = bpy.data.objects[Arm_Name].data.edit_bones['FaceRig.Root']
                            Head = bpy.data.objects[Arm_Name].data.edit_bones[data.Tables_Parent_Bone]
                        
                            Table.parent = Head
                            FaceRig.parent = Head
                            bpy.ops.object.mode_set(mode='OBJECT')
                        
                        clean()
                        

                        
                        
                        self.report({'INFO'}, 'Face Bones has been rigged with Shape Keys!')
            elif data.Face_Bones is True:
                for i in mod:
                        
                        # Runs if it finds the Armature modifier
                        if i.type == 'ARMATURE':
                            # Changes the armature name variable to the armature attached to the face
                            # global Arm_Name
                            Arm_Name = i.object.name
                            
                            # Sets up the Face rig
                            Build_FaceRig_action_constraint(bpy.data.objects[Arm_Name])
                            Build_Phoneme_action_constraint(bpy.data.objects[Arm_Name])
                            
                                                        
                            # Locks the tables to a bone if the user set it                        
                            if data.Tables_Parent_Bone != "":
                                bpy.ops.object.mode_set(mode='EDIT')
                                Table = bpy.data.objects[Arm_Name].data.edit_bones['Phoneme_Table.Root']
                                FaceRig = bpy.data.objects[Arm_Name].data.edit_bones['FaceRig.Root']
                                Head = bpy.data.objects[Arm_Name].data.edit_bones[data.Tables_Parent_Bone]
                            
                                Table.parent = Head
                                FaceRig.parent = Head
                                bpy.ops.object.mode_set(mode='OBJECT')
                            
                            clean()
                            self.report({'INFO'}, 'Face Bones has been rigged with Face Bones!')
                #lock_FaceRiggerComponents(data.Tables_Parent_Bone)
            
        else:
             self.report({'ERROR'}, 'Phoneme Table and FaceRig are not present!')

        return {'FINISHED'}
#==================================================================================================
# 
#==================================================================================================    






#==================================================================================================
# Mirror Select Operator
#================================================================================================== 
class FaceRigger_Mirror_Select(bpy.types.Operator):
    """Tooltip"""
    bl_idname = "bone.mirrorselect"
    bl_label = "FaceRigger : Mirror Select"


    def execute(self, context):
        # Stores the armature object and the selected pose bones
        arm = bpy.context.object.name
        bones = bpy.context.selected_pose_bones

        # Sets the lists of prefixes and sufixes 
        r_prefixes = ['R','r','R_','r_','R.','r.','Right','right','Right ',' right ']
        l_prefixes = ['L','l','L_','l_','L.','l.','Left','left','Left ',' left ']
        r_sufixes = [ 'R',' r','_R','_r','.R','.r','Right','right',' Right',' right ']
        l_sufixes = ['L',' l','_L','_l','.L','.l','Left','left',' Left',' left ']
        r_middle = [' Right ',' right ']
        l_middle = [' Left ',' left ']

        selection_list = list()

        # Iterates through the selected bones
        for i in bones:
            # Sets a temporary variable to store 
            temp_bone_name = ""
            # This is used to skip some for loops if it found the stuff
            found = False
            
            for r in r_prefixes:
                if i.name.startswith(r) and found is False:
                    temp_bone_name = i.name.replace(r,l_prefixes[r_prefixes.index(r)],1)
                    found = True
                    break
     #   if found is not True:
            for l in l_prefixes:
                if i.name.startswith(l) and found is False:
                    print("Found")
                    temp_bone_name = i.name.replace(l,r_prefixes[l_prefixes.index(l)],1)
                    found = True
                    break
      #  if found is not True:
            for l in l_sufixes:
                if i.name.endswith(l) and found is False:
                    print("Found 4")
                    temp_bone_name = i.name.replace(l,r_sufixes[l_sufixes.index(l)],len(l_sufixes)-1)
                    found = True
                    break
    #    if found is not True:
            for r in r_sufixes:
                if i.name.endswith(r) and found is False:
                    temp_bone_name = i.name.replace(r,l_sufixes[r_sufixes.index(r)],len(r_sufixes)-1)
                    found = True
                    break
#            if found is not True:
            for l in l_middle:
                if i.name.find(l) != -1 and found is False:
                    temp_bone_name = i.name.replace(l,r_middle[l_middle.index(l)],i.name.find(l))
                    found = True
                    break
#            if found is not True:
                for r in r_middle:
                    if i.name.find(r) != -1 and found is False:
                        temp_bone_name = i.name.replace(r,l_middle[r_middle.index(r)],i.name.find(r))
                        found = True
                        break
            if found is True:
                print(temp_bone_name)
                if bpy.data.objects[arm].data.bones.get(temp_bone_name) is not None:
                    print("setting")
                    bpy.data.objects[arm].data.bones[temp_bone_name].select = True
        return {'FINISHED'}    
#==================================================================================================
# 
#================================================================================================== 





#==================================================================================================
# Refactor Face Bones Function
#================================================================================================== 
def Resize_FaceBones(Arm_Name, Head, Layer,multiply):
    # Sets the mode to Edit
    bpy.ops.object.mode_set(mode='EDIT')                      
    
    # Sets the armature and head bone                       
    Armature_Name = Arm_Name
    Head_Bone = Head

    # Sets the reference scale and retrieves the children of the head bone                           
    Head_Scale = (bpy.data.objects[Armature_Name].data.edit_bones[Head_Bone].length)
    Head_Children = bpy.data.objects[Armature_Name].data.edit_bones[Head_Bone].children_recursive

    # Iterates through all of the children of the head bones
    for i in Head_Children:
        # Rotates the bone so it's faceing the screen
        bone = bpy.data.objects[Armature_Name].data.edit_bones[i.name]
        bone.tail = bone.head + Vector((0.0, -Head_Scale/10 , 0.0))

    # Sets the mode to pose and gets the children of the head bones again
    bpy.ops.object.mode_set(mode='POSE')
    Pose_Head_Children = bpy.data.objects[Armature_Name].data.bones[Head_Bone].children_recursive

    #Creates a layers list and set the layer that will be present
    new_layers = [False] * 32
    new_layers[Layer] = True
    
    # Iterates through all of the face bones
    for p in Pose_Head_Children:
        # Sets the currently selected bone custom bone and moves it to the desired bone layer
        bone = bpy.data.objects[Armature_Name].pose.bones[p.name]
        bone.custom_shape = bpy.data.objects['FaceRigger.Custom Face Bone']
        bpy.data.objects[Armature_Name].data.bones[p.name].show_wire = True
        bone.use_custom_shape_bone_size = False
        bone.custom_shape_scale_xyz = multiply * Vector((1.0, 1.0, 1.0))
        p.layers = new_layers
#==================================================================================================
# 
#================================================================================================== 





#==================================================================================================
# Refactor Face Bones Operator
#================================================================================================== 
class FaceBones_Refactor(bpy.types.Operator):
    bl_idname = "bones.refactor_facebones"
    bl_label = "FaceRigger: Refactor Face Bones"


    Bone_Layer: bpy.props.IntProperty(name="Int")
    Multiplier: bpy.props.FloatProperty(name="Int")
    Head_Bone: bpy.props.StringProperty(name="String Value")
    Arm_Name: bpy.props.StringProperty(name="String Value")
    

    def execute(self, context):
        if self.Head_Bone != "" and self.Bone_Layer >= 0 and self.Bone_Layer <= 31 and self.Multiplier > 0:
            Resize_FaceBones(self.Arm_Name,self.Head_Bone, self.Bone_Layer,self.Multiplier)
            self.report({"INFO"}, "FaceBones Have Been Refactored")
            return {'FINISHED'}
        else:
            if self.Head_Bone == "":
                self.report({"WARNING"}, "Please Put In The Head Bone")
            elif self.Bone_Layer >= 0 or self.Bone_Layer <= 31:
                self.report({"WARNING"}, "Please Enter A Bone Layer Between 0 and 31")
            elif self.Multiplier <= 0 :
                self.report({"WARNING"}, "Please Set Multiplier To A Value Greater Than 0")
            else:
                self.report({"WARNING"}, "An Expected Error Occured Please Try Again")
            return {"CANCELLED"} 
        

    def invoke(self, context, event):
        self.Multiplier = 1.00
        wm = context.window_manager
        return wm.invoke_props_dialog(self)

    def draw(self, context):
        self.Arm_Name = bpy.context.object.name
        Arm = bpy.data.objects[self.Arm_Name].data
        layout = self.layout
        col = layout.box()
        row = col.row()
        row.prop_search(self, "Head_Bone", Arm, "bones", text="Head Bone")
        row = col.row()
        row.prop(self, "Bone_Layer", text="Bone Layer")
        row = col.row()
        row.prop(self, "Multiplier", text="Bone Scale Multiplier")
#==================================================================================================
# 
#==================================================================================================        





#==================================================================================================
# Pose Mirror Operator
#================================================================================================== 
class FaceRigger_PoseMirror(bpy.types.Operator):
    """Tooltip"""
    bl_idname = "bones.faceriggermirror"
    bl_label = "FaceRigger : Mirror Pose"


    def execute(self, context):
        
        # Stores the armature object and the selected pose bones
        arm = bpy.context.object.name
        bones = bpy.context.selected_pose_bones

        # Sets the lists of prefixes and sufixes 
        r_prefixes = ['R','r','R_','r_','R.','r.','Right','right']
        l_prefixes = ['L','l','L_','l_','L.','l.','Left','left']
        r_sufixes = ['R',' r','_R','_r','.R','.r','Right','right']
        l_sufixes = ['L',' l','_L','_l','.L','.l','Left','left']
        r_middle = [' Right ',' right ']
        l_middle = [' Left ',' left ']

        selection_list = list()

        # Iterates through the selected bones
        for i in bones:
            # Sets a temporary variable to store 
            temp_bone_name = ""
            # This is used to skip some for loops if it found the stuff
            found = False
            
            for r in r_prefixes:
                if i.name.startswith(r):
                    temp_bone_name = i.name.replace(r,l_prefixes[r_prefixes.index(r)],1)
                    found = True
                    break
            #if found is not True:
            for l in l_prefixes:
                if i.name.startswith(l) and found is False:
                    temp_bone_name = i.name.replace(l,r_prefixes[l_prefixes.index(l)],1)
                    found = True
                    break
        #if found is not True:
            for l in l_sufixes:
                if i.name.endswith(l) and found is False:
                    temp_bone_name = i.name.replace(l,r_sufixes[l_sufixes.index(l)],len(l_sufixes)-1)
                    found = True
                    break
        #if found is not True:
            for r in r_sufixes:
                if i.name.endswith(r) and found is False:
                    temp_bone_name = i.name.replace(r,l_sufixes[r_sufixes.index(r)],len(r_sufixes)-1)
                    found = True
                    break
 #           if found is not True:
            for l in l_middle:
                if i.name.find(l) != -1 and found is False:
                    temp_bone_name = i.name.replace(l,r_middle[l_middle.index(l)],i.name.find(l))
                    found = True
                    break
     #       if found is not True:
            for r in r_middle:
                print("found")
                if i.name.find(r) != -1 and found is False:
                    temp_bone_name = i.name.replace(r,l_middle[r_middle.index(r)],i.name.find(r))
                    found = True
                    break
            if found is True:
                if bpy.data.objects[arm].data.bones.get(temp_bone_name) is not None:
                    target = bpy.context.object.pose.bones[temp_bone_name]
                    m = i.matrix.Diagonal((-1,1,1,0))
                    target.matrix = m @ i.matrix
                    target.scale = (1,1,1)

                    bpy.context.view_layer.update()

                    z = mathutils.Euler((0,0,math.radians(180))).to_matrix().to_4x4()
                    target.matrix = target.matrix @ z
                    bpy.context.view_layer.update()

                    y = mathutils.Euler((0,math.radians(180),0)).to_matrix().to_4x4()
                    target.matrix = target.matrix @ y
                    bpy.data.objects[arm].data.bones[temp_bone_name].select = True
    
        return {'FINISHED'}
#==================================================================================================
# 
#================================================================================================== 







#==================================================================================================
# Pose Menu Menu  
#================================================================================================== 
class FaceRigger_Pose_Menu(bpy.types.Menu):
    bl_label = "Face Rigger Pose Menu"
    bl_idname = "OBJECT_MT_facerigger_pose_menu"

    def draw(self, context):
        layout = self.layout
        layout.operator("bone.load_facerig")
        layout.operator("bone.mirrorselect")
        layout.operator("bones.faceriggermirror")
#==================================================================================================
# 
#================================================================================================== 






#==================================================================================================
# Pose Menu Operator
#================================================================================================== 
class FaceRigger_Operators(bpy.types.Panel):
    #BL Definations
    bl_idname = "POSE_PT_FACE"
    bl_label = "Face Rigger Operators"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = "Face Rigger Operators"


    def draw(self, context):
        layout = self.layout
        layout.operator("bone.load_facerig")
        layout.operator("bone.mirrorselect")
        layout.operator("bones.faceriggermirror")
        layout.operator_context = 'INVOKE_DEFAULT' #'INVOKE_AREA'
        layout.operator("bones.refactor_facebones")
#==================================================================================================
# 
#================================================================================================== 






#==================================================================================================
# FaceRigger Panel Class
#================================================================================================== 
class FaceRigger(bpy.types.Panel):
    #BL Definations
    bl_idname = "OBJECT_PT_Face"
    bl_label = "Face Rigger"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "object"
    
    
    def update(self, context):
         bpy.context.scene.Face.face_rig_data.Look_Up_SK = "test"
        
        
    #Draw Function. Handles the UI 
    def draw(self, context): 
        # Sets up the Layout Object a new box
        scene = context.scene
        layout = self.layout.box()
        
        # Create a new row for storing the label
        row = layout.row()
        row.alignment = "CENTER"
        row.label(text = "Face Mesh")
        
        # Field to store the mesh object
        row = layout.row()
        row.prop(scene, "Face", text ="")

        # Checks to see if the Face field is not empty        
        if bpy.context.scene.Face is not None:
            
            # Stores the current data to data variable
            data = bpy.context.scene.Face.face_rig_data
            
            mods = bpy.data.objects[bpy.context.scene.Face.name].modifiers
            
            #Sets the bone layer the Tables should appear in
            layout = self.layout.box()
            row = layout.row()
            row.alignment = "CENTER"
            row.label(text="Set the Tables' Active Layer")
            row = layout.row()
            row.alignment = "CENTER"
            row.prop(data, "Phoneme_Table_Bone_Layers")
            row.prop(data, "FaceRig_Bone_Layers")
            
            layout = self.layout.box()
            row = layout.row()
            row.alignment = "CENTER"
            row.label(text="Join To Armature?")
            row = layout.row()
            row.alignment = "CENTER"
            row.prop(data, "Join_To_Armature")

            # Looks through the modifiers attached to the mesh
            for m in mods:
                # Checks to see if the current modifier in the for loop is an armature modifier
                if m.type == 'ARMATURE':
                    
                    # Gets the armature object and its name
                    Arm_Name = bpy.data.objects[bpy.context.scene.Face.name].modifiers[m.name].object.name
                    Arm = bpy.data.objects[Arm_Name].data
                                        
                    layout = self.layout.box()
                    row = layout.row()
                    row.alignment = "CENTER"
                    row.label(text="Select A Bone To Lock The Tables To")
                    row = layout.row()
                    row.alignment = "CENTER"
                    row.prop_search(data,"Tables_Parent_Bone",Arm,"bones", text = "")
                    break
                    
            
            # Sets up the option to switch between Shape Keys or Face Bones
            layout = self.layout.box()
            row = layout.row()
            row.alignment = "CENTER"
            row.label(text="Select A Mode")
            row = layout.row()
            row.alignment = "CENTER"
            row.prop(data, "Shape_Keys")
            row.prop(data, "Face_Bones")
            
            
            if data.Shape_Keys is True and data.Face_Bones is True:
                    layout = self.layout.box()
                    row = layout.row()
                    row.alignment = "CENTER"
                    row.label(text = "Cannot select both!")
            
            elif data.Shape_Keys is True:
                if bpy.data.objects[bpy.context.scene.Face.name].data.shape_keys is not None:   
                    #Retrieves the shape keys              
                    mesh = bpy.data.objects[bpy.context.scene.Face.name].data.shape_keys
                                    
                    # Creates the Eye Target Field First
                    layout = self.layout.box()    
                    row = layout.row()
                    row.alignment = "CENTER"
                    row.label(text = "Eye Target")
                    
                   
                    col = layout.column()
                    row = col.row(align = True)
                    col = row.column()
                    col.alignment = "CENTER"
                    col.label(text = "Up")
                    col = row.column()
                    col.alignment = "CENTER"
                    col.label(text = "Down")
                    col = row.column()
                    col.alignment = "CENTER"
                    col.label(text = "Left")
                    col = row.column()
                    col.alignment = "CENTER"
                    col.label(text = "Right")
                    
                    row = layout.row(align = True)
                    row.prop_search(data, "Look_Up_SK", mesh ,"key_blocks" , text = "")
                    row.prop_search(data, "Look_Down_SK", mesh ,"key_blocks" , text = "")
                    row.prop_search(data, "Look_Left_SK", mesh ,"key_blocks" , text = "")
                    row.prop_search(data, "Look_Right_SK", mesh ,"key_blocks" , text = "")
                    
                    
                    layout = self.layout.box()    
                    row = layout.row()
                    row.label(text = "Upper Eyelids")
                    
                    col = layout.column()
                    row = col.row(align = True)
                    col = row.column()
                    col.alignment = "CENTER"
                    col.label(text = "Left")
                    col = row.column()
                    col.alignment = "CENTER"
                    col.label(text = "Right")
                    
                    row = layout.row(align = True)
                    row.prop_search(data, "Left_Blink", mesh ,"key_blocks" , text = "Blink")
                    row.prop_search(data, "Right_Blink", mesh ,"key_blocks" , text = "")
                    row = layout.row(align = True)
                    row.prop_search(data, "Left_Wide", mesh ,"key_blocks" , text = "Wide")
                    row.prop_search(data, "Right_Wide", mesh ,"key_blocks" , text = "")
                    
                    
                    layout = self.layout.box()    
                    row = layout.row()
                    row.label(text = "Lower Eyelids")
                    
                    col = layout.column()
                    row = col.row(align = True)
                    col = row.column()
                    col.alignment = "CENTER"
                    col.label(text = "Left")
                    col = row.column()
                    col.alignment = "CENTER"
                    col.label(text = "Right")
                    
                    row = layout.row(align = True)
                    row.prop_search(data, "Left_Squint", mesh ,"key_blocks" , text = "Squint")
                    row.prop_search(data, "Right_Squint", mesh ,"key_blocks" , text = "")
                    
                    layout = self.layout.box()    
                    row = layout.row()
                    row.label(text = "Eyebrows")

                    
                    col = layout.row()
                    col.label(text = "Left")
                    col.label(text = "Right")
                    
                    row = layout.row()
                    row.prop_search(data, "Left_Middle_Eyebrow_Up", mesh ,"key_blocks" , text = "Up")
                    row.prop_search(data, "Right_Middle_Eyebrow_Up", mesh ,"key_blocks" , text = "")
        
                
                    row = layout.row()
                    row.prop_search(data, "Left_Middle_Eyebrow_Down", mesh ,"key_blocks" , text = "Down")
                    row.prop_search(data, "Right_Middle_Eyebrow_Down", mesh ,"key_blocks" , text = "")
       
                    row = layout.row()
                    row.prop_search(data, "Left_Inner_Eyebrow_Up", mesh ,"key_blocks" , text = "Sad")
                    row.prop_search(data, "Right_Inner_Eyebrow_Up", mesh ,"key_blocks" , text = "")
                    
                    row = layout.row()
                  
                    row.prop_search(data, "Left_Inner_Eyebrow_Down", mesh ,"key_blocks" , text = "Angry")
                    row.prop_search(data, "Right_Inner_Eyebrow_Down", mesh ,"key_blocks" , text = "")
                    
                    
                    
                    layout = self.layout.box()    
                    row = layout.row()
                    row.label(text = "Lip Corners")

                    
                    col = layout.row()
                    col.label(text = "Left")
                    col.label(text = "Right")
                    
                    row = layout.row()
                    row.prop_search(data, "Left_Lipcorner_Up", mesh ,"key_blocks" , text = "Up")
                    row.prop_search(data, "Right_Lipcorner_Up", mesh ,"key_blocks" , text = "")
        
                
                    row = layout.row()
                    row.prop_search(data, "Left_Lipcorner_Down", mesh ,"key_blocks" , text = "Down")
                    row.prop_search(data, "Right_Lipcorner_Down", mesh ,"key_blocks" , text = "")
                    
                    
                    # Creates the Phoneme Library
                    layout = self.layout.box()    
                    row = layout.row()
                    row.label(text = "Phoneme")
                    
                   
                    col = layout.column()
                    row = col.row(align = True)
                    col = row.column()
                    col.alignment = "CENTER"
                    col.label(text = "Ah")
                    col = row.column()
                    col.alignment = "CENTER"
                    col.label(text = "Oh")
                    col = row.column()
                    col.alignment = "CENTER"
                    col.label(text = "Tndd")
                    col = row.column()
                    col.alignment = "CENTER"
                    col.label(text = "Bm")
                    col = row.column()
                    col.alignment = "CENTER"
                    col.label(text = "Fv")
                    
                    
                    row = layout.row(align = True)
                    row.prop_search(data, "Phoneme_Ah", mesh ,"key_blocks" , text = "")
                    row.prop_search(data, "Phoneme_Oh", mesh ,"key_blocks" , text = "")
                    row.prop_search(data, "Phoneme_Tn", mesh ,"key_blocks" , text = "")
                    row.prop_search(data, "Phoneme_Bm", mesh ,"key_blocks" , text = "")
                    row.prop_search(data, "Phoneme_Fv", mesh ,"key_blocks" , text = "")
                    
                    
                    # Adds the run operator
                    layout = self.layout
                    row = layout.row()
                    row.operator("object.simple_operator")
                else:
                    # Displays if the selected mesh has no shapekeys
                    layout = self.layout.box()
                    row = layout.row()
                    row.alignment = "CENTER"
                    row.label(text = "Select a mesh with shape keys!")
            # Runs if Face Rig Bones option is selected
            elif data.Face_Bones is True:           
                # Displays if the selected mesh has no shapekeys
                layout = self.layout.box()
                row = layout.row()
                row.alignment = "CENTER"
                row.label(text = "Eyes")
                row = layout.row()
                row.prop_search(data, "Eyes_UpDown", bpy.data ,"actions" , text = "Eyes Up/Down")
                row = layout.row()
                row.prop_search(data, "Eyes_LeftRight", bpy.data ,"actions" , text = "Eyes Left/Right")
 
 
                layout = self.layout.box()
                row = layout.row()
                row.alignment = "CENTER"
                row.label(text = "Eyelids")
                row = layout.row()
                row.prop_search(data, "Closedlid", bpy.data ,"actions" , text = "Upper Eyelids")
                row = layout.row()
                row.prop_search(data, "Lowerlid", bpy.data ,"actions" , text = "Lower Eyelids")
                
                
                layout = self.layout.box()
                row = layout.row()
                row.alignment = "CENTER"
                row.label(text = "Eyebrows")
                row = layout.row()
                row.prop_search(data, "BrowIn", bpy.data ,"actions" , text = "BrowIn")
                row = layout.row()
                row.prop_search(data, "Brow", bpy.data ,"actions" , text = "Brow Up/Down")
                #Lowerlid
                
                layout = self.layout.box()
                row = layout.row()
                row.alignment = "CENTER"
                row.label(text = "Lip Corners")
                row = layout.row()
                row.prop_search(data, "Lipcorner_UpDown", bpy.data ,"actions" , text = "Lip Up/Down")
                row = layout.row()
                row.prop_search(data, "Lipcorner_LeftRight", bpy.data ,"actions" , text = "Lip Left/Right")
            
                # Creates the Phoneme Library
                layout = self.layout.box()    
                row = layout.row()
                row.alignment = "CENTER"
                row.label(text = "Phoneme")
                
               
                col = layout.column()
                row = col.row(align = True)
                col = row.column()
                col.alignment = "CENTER"
                col.label(text = "Ah")
                col = row.column()
                col.alignment = "CENTER"
                col.label(text = "Oh")
                col = row.column()
                col.alignment = "CENTER"
                col.label(text = "Tn")
                col = row.column()
                col.alignment = "CENTER"
                col.label(text = "Bm")
                col = row.column()
                col.alignment = "CENTER"
                col.label(text = "Fv")
                
                
                row = layout.row(align = True)
                row.prop_search(data, "Phoneme_Ah", bpy.data ,"actions" , text = "")
                row.prop_search(data, "Phoneme_Oh", bpy.data ,"actions" , text = "")
                row.prop_search(data, "Phoneme_Tn", bpy.data ,"actions" , text = "")
                row.prop_search(data, "Phoneme_Bm", bpy.data ,"actions" , text = "")
                row.prop_search(data, "Phoneme_Fv", bpy.data ,"actions" , text = "")
            
                # Adds the run operator
                layout = self.layout
                row = layout.row()
                row.operator("object.simple_operator")
#==================================================================================================
# 
#================================================================================================== 
                


    
        
        

#==================================================================================================
# Register Classes (can replace with a more efficent one later)
#==================================================================================================        
def register():
    bpy.utils.register_class(FaceRiggerOp_LockComponents)
    bpy.utils.register_class(Load_FaceRigger_Components_Operator)
    bpy.utils.register_class(FaceRigger)
    bpy.utils.register_class(SimpleOperator)
    bpy.utils.register_class(FaceRigData)
    bpy.utils.register_class(FaceRigger_Pose_Menu)
    bpy.utils.register_class(FaceRigger_Mirror_Select)
    bpy.utils.register_class(FaceRigger_PoseMirror)
    bpy.utils.register_class(FaceRigger_Operators)
    bpy.utils.register_class(FaceBones_Refactor)
    
    # Creates a pointer property used in the UI as well as the FaceRig property
    bpy.types.Scene.Face = PointerProperty(type=bpy.types.Object , update = prefill)
    bpy.types.Object.face_rig_data = PointerProperty(type=FaceRigData)
#==================================================================================================
# 
#================================================================================================== 






#==================================================================================================
# Unregister Classes 
#================================================================================================== 
def unregister():
    bpy.utils.unregister_class(FaceRiggerOp_LockComponents)
    bpy.utils.unregister_class(Load_FaceRigger_Components_Operator)
    bpy.utils.unregister_class(FaceRigger)
    bpy.utils.unregister_class(SimpleOperator)
    bpy.utils.unregister_class(FaceRigData)
    bpy.utils.unregister_class(FaceRigger_Pose_Menu)
    bpy.utils.unregister_class(FaceRigger_Mirror_Select)
    bpy.utils.unregister_class(FaceRigger_PoseMirror)
    bpy.utils.unregister_class(FaceRigger_Operators)
    bpy.utils.unregister_class(FaceBones_Refactor)

    del bpy.types.Object.face_rig_data
#==================================================================================================
# 
#================================================================================================== 






#==================================================================================================
# Runs the code
#================================================================================================== 
if __name__ == "__main__":
    register()
#==================================================================================================
# 
#================================================================================================== 